# UlamMethod

<!-- [![Build Status](https://github.com/70Gage70/UlamMethod.jl/actions/workflows/CI.yml/badge.svg?branch=main)](https://github.com/70Gage70/UlamMethod.jl/actions/workflows/CI.yml?query=branch%3Amain) -->

## Introduction

This package is an implementation of Ulam's method [[1]](#1), [[2]](#2) (see also Galerkin projection [[3]](#3)) for the discretization of a stochastic operator. Given a set of two-dimensional, one-step trajectories $(x_{0, 1}, y_{0, 1}) \to  (x_{T, 1}, y_{T, 1}), (x_{0, 2}, y_{0, 2}) \to  (x_{T, 2}, y_{T, 2}) \dots$ defined in a rectangular domain, the following steps are performed

- A covering of the the domain is generated by polygons according to several different [algorithms.](#binning-algorithms)
- Bin the data according to the covering to generate a transition probability matrix such that any data outside the computational domain is lumped into a so-called "nirvana" state [[4]](#4).
- Transition path theory statistics [[5]](#5) are computed on the stationary, time-homogenous Markov chain induced by the transition probability matrix.

## Installation

In the Julia REPL, run the following code and follow the prompts:

```julia
using Pkg
Pkg.add("UlamMethod")
```

Make the functions in this package available to use in your code by including the following line:

```julia
using UlamMethod
```

## Usage

### Ulam's Method

Prepare the data into one of two formats: [.mat](https://github.com/JuliaIO/MAT.jl) or [.h5](https://github.com/JuliaIO/HDF5.jl) such that the head of the file contains four variables, `x0`, `xT`, `y0` and `yT`. See `test/x0x5-NA-undrogued.mat` and `test/x0x5-NA-undrogued.h5` for example trajectory data from undrogued drifters in the North Atlantic obtained from the NOAA GDP [[6]](#6). We will use the .mat file for this example

```julia
f_in = "x0x5-NA-undrogued.mat"
```

Then, select the size of the computational domain, that is, a rectangle of the form `[xmin, xmax, ymin, ymax]`. All data outside this rectangle, if any, will be added to the nirvana state.

```julia
corners = [-100, 15, -9, 39]
```

Next, select a [binning algorithm.](#binning-algorithms) We will use the `reg` algorithm here.

```julia
type = "reg"
```

Finally, choose the number of polygons in the covering.

```julia
n_polys = 500
```

We execute Ulam's method as follows

```julia
ulam = ulam_method(f_in, n_polys, type, corners)
```

The output of Ulam's method is a dictionary. The most relevant keys are `P_closed` which points to the full transition probability matrix with nirvana as the last row/column and `polys` which returns the polygons of the covering as a matrix. This matrix is a list of nodes such that the first column is the index of the node identifying the polyon it belongs to and the second and third columns are the x, y coordinates of that node. 

### Transition path theory

Currently, infinite-time transition path theory is supported. To compute transition path theory statistics, we first choose the coordinates defining the locations of the source $A$ and the target $B$. These can either be lists of points or a rectangular region using the function `AB_smear`. In this example, we'll set $A$ to be a single point and $B$ to be a region.

```julia
A_centers = [
    -18.0 17.0;
    ]
    
B_centers = AB_smear(-98.0, -92.0, 17.7, 32.0, resolution = 0.1)
```

Then, we compute the statistics as follows


```julia
tpt = tpt_from_ulam(ulam, A_centers, B_centers)
```

The output of `tpt_from_ulam` is a dictionary with all of the standard transition path theory statistics such as the reactive density `"muAB"`, transition time `"tAB"` and effective flux `"f+"`.


## Binning algorithms

- `"reg"` A covering of the domain by a grid of squares.
- `"hex"` A covering of the domain by a grid of hexagons.
- `"vor"` A covering of the domain by a [Voronoi tesselation](https://en.wikipedia.org/wiki/Voronoi_diagram) based on clustering the observational data using [k-means](https://en.wikipedia.org/wiki/K-means_clustering). Since the data are clustered, this algorithm is significantly slower than the regular coverings.

## References

<a id = "1">[1]</a> Ulam, Stanislaw M. A collection of mathematical problems. No. 8. Interscience Publishers, 1960.

<a id = "2">[2]</a> Li, Tien-Yien. "Finite approximation for the Frobenius-Perron operator. A solution to Ulam's conjecture." Journal of Approximation theory 17.2 (1976): 177-186.

<a id = "3">[3]</a> Reddy, Junuthula Narasimha. Introduction to the finite element method. McGraw-Hill Education, 2019.

<a id = "4">[4]</a> Miron, Philippe, et al. "Transition paths of marine debris and the stability of the garbage patches<? A3B2 show [editpick]?>." Chaos: An Interdisciplinary Journal of Nonlinear Science 31.3 (2021): 033101.

<a id = "5">[5]</a>  Vanden-Eijnden, Eric. "Transition path theory." Computer Simulations in Condensed Matter Systems: From Materials to Chemical Biology Volume 1. Springer, Berlin, Heidelberg, 2006. 453-493.

<a id = "6">[6]</a> Lumpkin, Rick, and Mayra Pazos. "Measuring surface currents with Surface Velocity Program drifters: the instrument, its data, and some recent results." Lagrangian analysis and prediction of coastal and ocean dynamics 39 (2007): 67.
